"""Mono X Objects"""

##pylint: disable=too-few-public-methods
class MonoXResponseType:
    """The baseline MonoX State class"""

    status: str = "error/offline"

    def print(self):
        """Print the MonoXResponse. Should be overridden by anything which implements this class."""
        return "Status: " + self.status


class MonoXFileEntry(MonoXResponseType):
    """A file entry consisting of an internal and external listing"""

    def __init__(self, internal_name: str, external_name: str) -> None:
        self.external = internal_name
        self.internal = external_name
        self.status = "file"

    def print(self):
        print(self.internal + ": " + self.external)


class FileList(MonoXResponseType):
    """handles lists of files"""

    def __init__(self, data: MonoXFileEntry) -> None:
        self.files = []
        self.status = "getfile"

        for field in data:
            if field in data[0] or data[-1]:
                continue  # not interested in packet open/close portion.
            split = field.split("/")
            self.files.append(MonoXFileEntry(split[0], split[1]))

    files = [MonoXFileEntry]

    def print(self):
        for file in self.files:
            file.print()


class InvalidResponse(MonoXResponseType):
    """Used when no response is provided"""

    def __init__(self, message) -> None:
        self.status = message

    def print(self):
        print("Invalid Response: " + self.status)


class SimpleResponse(MonoXResponseType):
    """Used when no response is provided"""

    def __init__(self, message) -> None:
        self.status = message

    def print(self):
        print("Response: " + self.status)


class MonoXSysInfo(MonoXResponseType):
    """The sysinfo handler. Handles sysinfo messages.

    eg message.
        sysinfo
        sysinfo,Photon Mono X 6K,V0.2.2,0000170300020034,SkyNet,end
    """

    def __init__(self, model="", firmware="", serial="", wifi="") -> None:
        self.model = model
        self.firmware = firmware
        self.serial = serial
        self.wifi = wifi
        self.status = "updated"

    def print(self):
        print("model: " + self.model)
        print("firmware: " + self.firmware)
        print("serial: " + self.serial)
        print("wifi: " + self.wifi)


class MonoXStatus(MonoXResponseType):  # pylint: disable=too-many-instance-attributes
    """Status object for MonoX.

    eg message.
        getstatus
        getstatus,stop
    """

    def __init__(self, message) -> None:

        self.status = message[1]
        if len(message) > 3:
            self.file = message[2]
        if len(message) > 4:
            self.total_layers = message[3]
        if len(message) > 5:
            self.layers_remaining = message[4]
        if len(message) > 6:
            self.current_layer = message[5]
        if len(message) > 7:
            self.seconds_elapse = message[6]
        if len(message) > 8:
            self.seconds_remaining = message[7]
        if len(message) > 9:
            self.total_volume = message[8]
        if len(message) > 10:
            self.mode = message[9]
        if len(message) > 11:
            self.unknown1 = message[10]
        if len(message) > 12:
            self.layer_height = message[11]
        if len(message) > 13:
            self.unknown2 = message[12]

    def print(self):
        print("status: " + self.status)
        if hasattr(self, "file"):
            print("file: " + self.file)
            print("total_layers: " + str(self.total_layers))
            print("layers_remaining: " + str(self.layers_remaining))
            print("current_layer: " + str(self.current_layer))
            print("seconds_remaining: " + str(self.seconds_remaining))
            print("total_volume: " + str(self.total_volume))
            print("mode: " + self.mode)
            print("unknown1: " + str(self.unknown1))
            print("layer_height: " + str(self.layer_height))
            print("unknown2: " + str(self.unknown2))


class MonoXPreviewImage(MonoXResponseType):
    """A file entry consisting of an internal and external listing"""

    def __init__(self, file_path: str) -> None:
        super().__init__()
        self.file_path = file_path
        self.status = "preview image"

    def print(self):
        print(f"preview located at {self.file_path}")



"""Communications with MonoX devices. -Adam Outler - email: adamoutler(a)hackedyour.info
The development environment is Visual Studio Code. See launch.json for auto-config.
"""


from asyncio.log import logger
from datetime import datetime
import os
import sys
from queue import Empty
import socket
import tempfile

from .response import (
    FileList,
    InvalidResponse,
    MonoXPreviewImage,
    MonoXResponseType,
    MonoXStatus,
    MonoXSysInfo,
)


# Port to listen on

HOST = "192.168.1.254"
COMMAND = "getstatus"
endRequired = ["goprint", "gostop", "gopause", "delfile"]
END = ",end"
ENCODING = "utf-8"
_LOGGER = logger
Any = object()


class UartWifi:  # pylint: disable=too-few-public-methods
    """Mono X Class"""

    def __init__(self, ip_address: str, port: int) -> None:
        self.server_address = (ip_address, port)
        self.raw = False
        self.telnet_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    def send_request(self, message_to_be_sent: str) -> MonoXResponseType:
        """sends the Mono X request."""
        request: str = bytes(message_to_be_sent, "utf-8")
        received: str = _do_request(self.telnet_socket, self.server_address, request)
        if self.raw:
            return received
        processed: MonoXResponseType = _do_handle(received)
        return processed


def _do_request(
    sock: socket, socket_address: tuple, to_be_sent: bytes
) -> MonoXResponseType:
    """Perform the request

    :param sock: the socket to use for the request
    :param request: the request to send
    :return: a MonoX object"""

    try:
        _LOGGER.debug("connecting to %s", socket_address)
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect(socket_address)
        sock.sendall(to_be_sent)
        if str(to_be_sent).startswith("b'getPreview2"):
            text_received = bytearray()
            print(text_received)
            end_time = datetime.now().microsecond + 10000
            while (
                not str(text_received).endswith(END)
                and datetime.now().microsecond < end_time
            ):
                text_received.extend(sock.recv(1))
        else:
            text_received = ""
            while not str(text_received).endswith(END):
                text_received += str(sock.recv(1).decode())
    except OSError as exception:
        _LOGGER.error(
            "Could not connect to AnyCubic printer at %s %s:%s",
            socket_address,
            exception.__class__,
            exception.__cause__,
        )
        return None
    finally:
        sock.close()
    return text_received


def __do_preview2(received_message: bytearray()):
    """Handles preview by writing to file."""

    filename = received_message.decode("utf_8").split(",", 3)[1]
    file = tempfile.gettempdir() + os.path.sep + filename + ".bmp"
    print(file)

    output_file = open(file=file, mode="rb")

    width = 240
    height = 168

    file_size = os.path.getsize(file)
    pos_in_image = 0
    max_line_length = width * 2
    slices = []
    for row in range(0, height):
        current_slice = bytearray()
        for current_byte in range(0, max_line_length):
            current_byte = (row * max_line_length) + current_byte
            if file_size >= current_byte:
                current_slice.append(output_file.read(2))
            current_byte += 2
            slices.append(current_slice)

    my_slice = [[]]
    for byte in slices:
        print(type(byte))
        my_slice[pos_in_image] = byte[0]
        pos_in_image += 1
        if pos_in_image > (240 * 2):
            pos_in_image = 0
            current_slice = bytearray
            slices.append(current_slice)

    # image = Image.new("RGB", (width, height))
    # file_format = "bmp"  # The file extension of the sourced data
    print(len(slices))

    print(slice)

    # bytes(byte_array)
    # image.write(output_file,file_format)

    # output_file.close()

    return MonoXPreviewImage("")


def _do_handle(message: Any):
    """Perform handling of the message received by the request"""
    if message is None:
        return "no response"

    lines = message.split("end")
    recognized_response: MonoXResponseType = None
    for line in lines:
        fields = line.split(",")
        message_type = fields[0]
        if len(fields) is Empty or len(fields) < 2:
            print("Done", file=sys.stderr)
            continue
        if len(fields) < 3:
            print(
                f"""Connection established, but no usable data was received.
            Transcript:\n{line}""",
                file=sys.stderr,
            )
            return ""
        if message_type == "getstatus":
            recognized_response = __do_status(fields)
        elif message_type == "getfile":
            recognized_response = __do_files(fields)
        elif message_type == "sysinfo":
            recognized_response = __do_sys_info(fields)
        elif message_type == "gethistory":
            recognized_response = __do_get_history(fields)
        elif message_type == "doPreview2":
            recognized_response = __do_preview2(fields)
        # goprint,49.pwmb,end
        elif message_type in ["goprint", "gostop", "gopause", "getmode", "getwifi"]:
            recognized_response = fields[1]
        else:
            print("unrecognized command: " + message_type, file=sys.stderr)
            print(line, file=sys.stderr)
        if recognized_response is not None:
            return recognized_response
        return InvalidResponse(line)


def __do_get_history(fields):
    """Handles history processing."""
    items = []
    for field in fields:
        if field in fields[0] or fields[-1]:
            continue
        items.append(field)
    return items


def __do_sys_info(fields):
    """Handles system info processing."""
    sys_info = MonoXSysInfo()
    if len(fields) > 2:
        sys_info.model = fields[1]
    if len(fields) > 3:
        sys_info.firmware = fields[2]
    if len(fields) > 4:
        sys_info.serial = fields[3]
    if len(fields) > 5:
        sys_info.wifi = fields[4]
    return sys_info


def __do_files(fields):
    """Handles file processing."""
    files = FileList(fields)
    return files


def __do_status(fields):
    """Handles status processing."""
    status = MonoXStatus(fields)
    return status



#! python3
"""Fake Anycubic Printer for tests"""
import getopt
import socket
import sys
import time


class AnycubicSimulator:
    """ "Simulator for Anycubic Printer."""

    ip_address = "127.0.0.1"
    port = "6000"
    printing = False
    serial = "0000170300020034"

    def __init__(self, ip_address: str, the_port: int) -> None:
        self.host = ip_address
        self.port = the_port
        self.printing = False
        self.serial = "234234234"
        self.my_socket:socket

    def sysinfo(self) -> str:
        """return sysinfo type"""
        return (
            "sysinfo,Photon Mono X 6K,V0.2.2,"
            + self.serial
            + ",SkyNet,endgetstatus,stop"
        )

    @staticmethod
    def getfile() -> str:
        """return getfile type"""
        return "getfile,Widget.pwmb/0.pwmb,end"

    def getstatus(self) -> str:
        """return getstatus type"""
        if self.printing:
            return (
                "getstatus,print,Phone stand hollow supported.pwmb"
                "/46.pwmb,2338,88,2062,51744,6844,~178mL,UV,39.38,0.05,0,end"
            )
        return "getstatus,stop\r\n,end"

    def goprint(self) -> str:
        """Do printing"""
        if self.printing:
            return "goprint,ERROR1,end"
        self.printing = True
        return "goprint,OK,end"

    def gostop(self) -> str:
        """Do printing"""
        if not self.printing:
            return "gostop,ERROR1,end"
        self.printing = False
        return "gostop,OK,end"

    def start_server(self):
        """Start the uart_wifi simualtor server"""
        self.my_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.my_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.my_socket.bind((self.ip_address, self.port))
        self.my_socket.listen(5)
        while True:
            try:
                conn, addr = self.my_socket.accept()
                data_received = ""
                while not data_received.endswith("\n"):
                    with conn:
                        print(f"Connected by {addr}")
                        data = conn.recv(1024)
                        if not data:
                            break
                        try:
                            decoded_data = data.decode()
                            print(decoded_data)
                        except UnicodeDecodeError:
                            continue
                        data_received += decoded_data
                        if data.endswith(("".encode(), "\n".encode())):
                            if data_received.startswith("getstatus"):
                                conn.sendall(self.getstatus().encode())
                            if data_received.startswith("sysinfo"):
                                conn.sendall(self.sysinfo().encode())
                            if data_received.startswith("getfile"):
                                conn.sendall(self.getfile().encode())
                            if data_received.startswith("goprint"):
                                conn.sendall(self.goprint().encode())
                            if data_received.startswith("gostop"):
                                conn.sendall(self.gostop().encode())
                            if data_received.startswith("getmode"):
                                conn.sendall("getmode,0,end".encode())
                            if (data_received.startswith("shutdown")):
                                sys.exit()
            except Exception:  # pylint: disable=broad-except
                pass
            finally:
                time.sleep(1)


opts, args = getopt.gnu_getopt(sys.argv, "i:p:", ["ipaddress=", "port="])

IP = "0.0.0.0"
PORT = 6000
for opt, arg in opts:
    if opt in ("-i", "--ipaddress"):
        IP = arg
    elif opt in ("-p", "--port"):
        PORT = arg
        print(arg)
AnycubicSimulator(IP, int(PORT)).start_server()



"""Main executable"""
import os
import getopt
import sys

from uart_wifi.communication import UartWifi
from uart_wifi.response import MonoXResponseType

PORT = 6000
HELP = (
    __file__
    + """
monox.py | Adam Outler (monox@hackedyour.info) | GPLv3

Usage: monox.py -i <ip address> -c <command>
args:
-i [--ipaddress=] - The IP address which your Anycubic Mono X can be reached

-c [--command=] - The command to send.
Commands may be used one-at-a-time. Only one command may be sent and it is
        expected to be in the format below.
    Command: getstatus - Returns a list of printer statuses.
    Command: getfile - returns a list of files in format <internal name>: <file name>.
        When referring to the file via command, use the <internal name>.
    Command: sysinfo - returns Model, Firmware version, Serial Number, and wifi network.
    Command: getwifi - displays the current wifi network name.
    Command: gopause - pauses the current print.
    Command: goresume - ends the current print.
    Command: gostop,end - stops the current print.
    Command: delfile,<internal name>,end - deletes a file.
    command: gethistory,end - gets the history and print settings of previous prints.
    Command: delhistory,end - deletes printing history.
    Command: goprint,<internal name>,end - Starts a print of the requested file
    Command: getPreview1,<internal name>,end - returns a list of dimensions used for the print.

Not Supported Commands may return unusable results.
    Command (Not Supported): getPreview2,<internal name>,end
    - returns a binary preview image of the print.

Unknown Commands are at your own risk and experimentation.
No attempt is made to process or stop execution of these commands.
    Command: detect
    Command: stopUV - unknown
    Command: getpara - unknown
    Command: getmode - unknown
    Command: setname - unknown
    Command: getname - unknown
    Command: setwifi - unknown
    Command: setZero - unknown
    Command: setZhome - unknown
    Command: setZmove - unknown
    Command: setZstop - unknown
    """
)
if __name__ == "__main__":
    os.system("script2.py", sys.argv)

    try:
        opts, args = getopt.gnu_getopt(sys.argv, "hi:c:", ["ipaddress=", "command="])
    # pylint: disable=broad-except
    except Exception:
        print(HELP)
        sys.exit(0)

    for opt, arg in opts:
        if opt == "-h":
            print(HELP)
            sys.exit()
        elif opt in ("-i", "--ipaddress"):
            HOST = arg
        elif opt in ("-c", "--command"):
            COMMAND = arg
            print(arg)

    response = UartWifi(HOST, PORT).send_request(COMMAND)

    if isinstance(response, MonoXResponseType):
        response.print()
    else:
        print(response)
